#ifndef __MANAAGER_H__
#define __MANAGER_H__

#include "shmem.h"
#include "sockio.h"
#include "pgqueue.h"

struct haObjectHeader { 
    unsigned size;
    unsigned cid;

    enum {	
	CID_UDT, 
	CID_FREE,
	CID_BITMAP
    };
};  

struct haStorageFreeBlock : public haObjectHeader { 
    haStorageFreeBlock* next;
};       

struct haStoragePageMap : haObjectHeader { 
    //
    // All pages except first occupied by large object are marked with non-zero
    // bit in this bitmap. Bits correspond to all other pages are zero.
    //
    char bits[1];
};

struct haStorageHeader : haFileHeader { 
    void*               root;
    haStoragePageMap*   pageMap;
    haStorageFreeBlock* freePageChain;
    haStorageFreeBlock* freeBlockChain[13];
};

class haManager {  
  public:
    bool open(char const* fileName, char const* host, int localPort, int remotePort, void* baseAddr = NULL);
    void commit();
    void close();

    void setRootObject(void* obj) { 
	hdr->root = obj; 
    }

    void* getRootObject() const { 
	return hdr->root; 
    }

    static haManager* getCurrent() { 
	return chain;
    }

    void* allocate(int cid, size_t size) 
    {
	if (size + sizeof(haObjectHeader) <= allocPageSize/2) { 
	    int n = blockChain[((size + sizeof(haObjectHeader) + 7) >> 3) - 1];
	    haStorageFreeBlock* bp = hdr->freeBlockChain[n];
	    if (bp != NULL) { 
		hdr->freeBlockChain[n] = bp->next;
		bp->size = size;
		bp->cid = cid;
		return (haObjectHeader*)bp+1;
	    }
	}
	return allocPage(cid, size);	
    }

    void  deallocate(void* obj) 
    { 
	haObjectHeader* hp = getHeader(obj);
	assert(hp->cid != haObjectHeader::CID_FREE);
	if (hp->size + sizeof(haObjectHeader) <= allocPageSize/2) {
	    int n = blockChain[((hp->size+sizeof(haObjectHeader)+7) >> 3) - 1];
	    haStorageFreeBlock* bp = (haStorageFreeBlock*)hp;
	    bp->next = hdr->freeBlockChain[n];
	    bp->cid = haObjectHeader::CID_FREE;
	    hdr->freeBlockChain[n] = bp;
	} else { 
	    freePage(hp);
	}
    }	

    haManager(size_t maxSize);

    static haManager* getManager(char* address) { 
	for (haManager* mng = chain; mng != NULL; mng = mng->next) { 
	    if ((size_t)(address - mng->shmem.base()) < mng->shmem.size()) { 
		return mng;
	    }
	}
	return NULL;
    }

    static bool pageFaultHandler(char* address);

  private:
    static void THREAD_PROC acceptProc(void* arg);
    static void THREAD_PROC updateProc(void* arg);

    enum { 
	allocPageSize = 512,   // bytes
	initPageMapSize = 8    // pages
    };

    void initialize();
    void handlePageFault(size_t offs);
    void recovery();
    void update();

    void* allocPage(int cid, size_t size);
    void  freePage(haObjectHeader* hp);

    haObjectHeader* getHeader(void* obj);

    size_t              maxSize;
    int*                dirtyPageMap;
    haSocket*           standbySock;
    haSocket*           prevStandbySock; // socket whcih was opened during last commit
    haSharedMemory      shmem;
    haPageQueue         queue;
    haThread            updateThread;
    haThread            acceptThread;
    int                 port;
    size_t              pageSize;
    int                 nDirtyPages;
    bool                acceptConnections;
    haStorageHeader*    hdr;
    haStorageFreeBlock* currFreePage;
    haManager*          next;
    static haManager*   chain;
    static const int blockSize[allocPageSize/16];
    static const int blockChain[allocPageSize/16];

};
	
#endif







